# -*- coding: utf-8 -*-
import sys
from StringIO import StringIO
import unittest
from skoolkittest import SkoolKitTestCase
from skoolkit.skoolctl import CtlWriter

DIRECTIVES = 'bcgituwz'

TEST_SKOOL = 'test-ctl.skool'

TEST_CTL = """; @org:32768=32768
; @start:32768
; @ignoredua:32768
c 32768 Routine
D 32768 Routine description
R 32768 A Some value
R 32768 B Another value
; @label:32768=START
  32768,1 Do nothing
; @bfix:32769=DEFB 2,3
B 32769,2,2 1-line B sub-block
; @nowarn:32771
; @isub:32772=DEFB 0,1
B 32771,3,1,2 2-line B sub-block
D 32774 Mid-block comment
; @ignoremrcua:32774
T 32774,5,5 T sub-block
; @keep:32779
; @ignoreua:32779
W 32779,2 W sub-block
Z 32781,2 Z sub-block
M 32783,12 Sub-block with instructions of various types
; @nolabel:32783
; @ofix:32785=DEFB 2
B 32785,1
; @rsub:32786=DEFW 0,1,2
W 32786,4,4
; @ssub:32790=DEFM "Lo"
T 32790,2,2
Z 32792,3
E 32768 End comment paragraph 1.
E 32768 End comment paragraph 2.
i 32796 Ignore block
b 49152 Data block
; @rem:49152=Hello!
g 49153 Game status buffer entry
W 49153
t 49155 Message
  49155,2,2
u 49157 Unused block
w 49158 Word block
  49158,4,2
z 49162 Zero block
; @end:49162
i 49172 Another ignore block""".split('\n')

TEST_CTL_BS = """c 32768
B 32769,2,2
B 32771,3,1,2
T 32774,5,5
W 32779,2
Z 32781,2
B 32785,1
W 32786,4,4
T 32790,2,2
Z 32792,3
i 32796
b 49152
g 49153
W 49153
t 49155
  49155,2,2
u 49157
w 49158
  49158,4,2
z 49162
i 49172""".split('\n')

TEST_CTL_BSC = """c 32768
  32768,1 Do nothing
B 32769,2,2 1-line B sub-block
B 32771,3,1,2 2-line B sub-block
T 32774,5,5 T sub-block
W 32779,2 W sub-block
Z 32781,2 Z sub-block
M 32783,12 Sub-block with instructions of various types
B 32785,1
W 32786,4,4
T 32790,2,2
Z 32792,3
i 32796
b 49152
g 49153
W 49153
t 49155
  49155,2,2
u 49157
w 49158
  49158,4,2
z 49162
i 49172""".split('\n')

class CtlWriterTest(SkoolKitTestCase):
    def _get_ctl(self, elements='btdrmsc', write_hex=False, write_asm_dirs=True):
        writer = CtlWriter(TEST_SKOOL, elements, write_hex, write_asm_dirs)
        out = StringIO()
        writer.write(out)
        return out.getvalue().split('\n')[:-1]

    def _assert_ctls_equal(self, actual_ctl, expected_ctl):
        self.assertEqual(len(actual_ctl), len(expected_ctl))
        for i, line in enumerate(actual_ctl):
            self.assertEqual(line, expected_ctl[i])

    def test_default_elements(self):
        self._assert_ctls_equal(self._get_ctl(), TEST_CTL)

    def test_default_elements_hex(self):
        ctl = self._get_ctl(write_hex=True)
        self.assertEqual(len(ctl), len(TEST_CTL))
        for i, line in enumerate(ctl):
            if not line.startswith(';'):
                hex_addr = '$%04X' % int(TEST_CTL[i][2:7])
                self.assertEqual(line[2:7], hex_addr)

    def test_default_elements_no_asm_dirs(self):
        ctl = self._get_ctl(write_asm_dirs=False)
        test_ctl = [line for line in TEST_CTL if not line.startswith(';')]
        self._assert_ctls_equal(ctl, test_ctl)

    def test_wb(self):
        ctl = self._get_ctl('b', write_asm_dirs=False)
        test_ctl = [line[:7] for line in TEST_CTL if line[0] in DIRECTIVES]
        self._assert_ctls_equal(ctl, test_ctl)

    def test_wbd(self):
        ctl = self._get_ctl('bd', write_asm_dirs=False)
        test_ctl = []
        for line in TEST_CTL:
            if line[0] in DIRECTIVES:
                test_ctl.append(line[:7])
            elif line.startswith('D 32768'):
                test_ctl.append(line)
        self._assert_ctls_equal(ctl, test_ctl)

    def test_wbm(self):
        ctl = self._get_ctl('bm', write_asm_dirs=False)
        test_ctl = []
        for line in TEST_CTL:
            if line[0] in DIRECTIVES:
                test_ctl.append(line[:7])
            elif line[0] in 'DE' and not line.startswith('D 32768'):
                test_ctl.append(line)
        self._assert_ctls_equal(ctl, test_ctl)

    def test_wbr(self):
        ctl = self._get_ctl('br', write_asm_dirs=False)
        test_ctl = []
        for line in TEST_CTL:
            if line[0] in DIRECTIVES:
                test_ctl.append(line[:7])
            elif line[0] == 'R':
                test_ctl.append(line)
        self._assert_ctls_equal(ctl, test_ctl)

    def test_wbs(self):
        ctl = self._get_ctl('bs', write_asm_dirs=False)
        self._assert_ctls_equal(ctl, TEST_CTL_BS)

    def test_wbsc(self):
        ctl = self._get_ctl('bsc', write_asm_dirs=False)
        self._assert_ctls_equal(ctl, TEST_CTL_BSC)

    def test_wbt(self):
        ctl = self._get_ctl('bt', write_asm_dirs=False)
        test_ctl = [line for line in TEST_CTL if line[0] in DIRECTIVES]
        self._assert_ctls_equal(ctl, test_ctl)

if __name__ == '__main__':
    unittest.main()
